@@ -26,6 +26,7 @@ import sys
 import pytest
 import asyncio
 from pathlib import Path
+from tempfile import TemporaryDirectory
 from aiohttp import ClientSession
 from aiohttp_socks import ProxyConnector
 
@@ -34,7 +35,7 @@ PACKAGE_PATH = str(Path(__file__).absolute().parent.parent)
 sys.path.append(PACKAGE_PATH)
 
 
-from tiny_gnupg import GnuPG, run
+from tiny_gnupg import *
 
 
 legacy_key = """-----BEGIN PGP PUBLIC KEY BLOCK-----
@@ -143,51 +144,51 @@ PzNx/ZogrbmOMOWsTCC1DFycAjCRtueMDG83FVl0zdX0fFR5cZnRdxde+Y6oyZ6u
 """
 
 
-_homedir = Path(PACKAGE_PATH).absolute() / "tiny_gnupg/gpghome"
-_options = _homedir / "gpg2.conf"
-_executable = "/usr/bin/gpg2"
-
-
 @pytest.fixture(scope="module")
 def gpg():
+    global _homedir, _options, _executable, main_passphrase
     print("setup".center(15, "-"))
 
     username = "testing_user"
-    email = "testing_user@testing.testing"
-    passphrase = "test_passphrase"
+    email_address = "testing_user@testing.testing"
+    main_passphrase = "test_passphrase"
+
+    tempdir = TemporaryDirectory("testing_tiny_gnupg")
+    _homedir = Path(tempdir.name).absolute()
+    _options = GnuPGConfig._DEFAULT_OPTIONS_PATH
+    _executable = GnuPGConfig._DEFAULT_EXECUTABLE_PATH
 
     gpg = GnuPG(
-        email=email,
+        email_address=email_address,
         username=username,
-        passphrase=passphrase,
+        passphrase=main_passphrase,
         homedir=_homedir,
         options=_options,
         executable=_executable,
     )
 
-    yield gpg
-    print("teardown".center(18, "-"))
-
-
-def pop(dictionary):
-    return next(iter(dictionary))
+    try:
+        yield gpg
+    finally:
+        print("teardown".center(18, "-"))
+        tempdir.cleanup()
 
 
 async def async_method_runner(gpg):
-    connector = gpg.network.Connector()
+    connector = gpg.keyserver.network.Connector()
     assert connector.__class__ == ProxyConnector
-    async with gpg.network.Session(connector=connector) as session:
-        session.__class__ == ClientSession
+    async with gpg.keyserver.network.Session() as session:
+        assert session.__class__ == ClientSession
 
 
 def test_instance(gpg):
     while True:
         try:
-            gpg.delete(gpg.email)
+            gpg.delete(gpg.user.email_address)
         except:
             break
     try:
-        gpg._set_homedir_permissions("/ridiculous_root_directory_not_real")
+        gpg.config.set_homedir("/ridiculous_root_directory_not_real")
     except FileNotFoundError:
         "Successfully failed to change permissions on an invalid dir. "
         "We refrain from testing on, or making, a priveleged folder "
@@ -198,44 +199,45 @@ def test_instance(gpg):
             "permissions."
         )
 
-    gpg.gen_key()
-    test_gpg = GnuPG(
-        email=gpg.user.email,
+    gpg.generate_key()
+    user = User(
+        email_address=gpg.user.email_address,
         username=gpg.user.username,
-        passphrase=gpg.user.passphrase,
+        passphrase=main_passphrase,
+    )
+    config = GnuPGConfig(
         homedir=_homedir,
         options=_options,
         executable=_executable,
     )
-    test_gpg.gen_key()
+    test_gpg = BaseGnuPG(user, config=config)
+    test_gpg.generate_key()
     assert gpg.user.username == test_gpg.user.username
-    assert gpg.user.email == test_gpg.user.email
+    assert gpg.user.email_address == test_gpg.user.email_address
     assert gpg.user.passphrase == test_gpg.user.passphrase
-    assert gpg.network.port == test_gpg.network.port
-    assert gpg.network.tor_port == test_gpg.network.tor_port
-    assert gpg.homedir == test_gpg.homedir
-    assert gpg.executable == test_gpg.executable
+    assert gpg.keyserver.network.port == test_gpg.keyserver.network.port
+    assert gpg.keyserver.network.tor_port == test_gpg.keyserver.network.tor_port
+    assert gpg.config.homedir == test_gpg.config.homedir
+    assert gpg.config.executable == test_gpg.config.executable
     run(async_method_runner(gpg))
-    assert gpg._search_prefix == test_gpg._search_prefix
-    assert gpg._keyserver == test_gpg._keyserver
-    assert str(gpg.network.port) in gpg._keyserver
-    assert gpg._keyserver_export_api == test_gpg._keyserver_export_api
-    assert gpg._keyserver_verify_api == test_gpg._keyserver_verify_api
-    assert gpg._searchserver == test_gpg._searchserver
+    assert gpg.keyserver._search_prefix == test_gpg.keyserver._search_prefix
+    assert gpg.keyserver._hostname == test_gpg.keyserver._hostname
+    assert str(gpg.keyserver.network.port) in gpg.keyserver._hostname
+    assert gpg.keyserver._search_template == test_gpg.keyserver._search_template
     assert gpg._base_command == test_gpg._base_command
     assert gpg._base_passphrase_command == test_gpg._base_passphrase_command
     ###
-    assert gpg.network.port == 80
-    assert gpg.network.tor_port == 9050
-    assert ".onion" in gpg._keyserver
-    assert ".onion" in gpg._searchserver
+    assert gpg.keyserver.network.port == 80
+    assert gpg.keyserver.network.tor_port == 9050
+    assert ".onion" in gpg.keyserver._hostname
+    assert ".onion" in gpg.keyserver._search_template
     assert len(gpg.fingerprint) == 40
     assert type(gpg.fingerprint) == str
-    assert gpg.homedir.endswith("gpghome")
+    assert gpg.config.homedir.endswith("testing_tiny_gnupg")
     assert gpg.user.username == "testing_user"
-    assert gpg.executable.endswith("gpg2")
-    assert gpg.user.passphrase == "test_passphrase"
-    assert gpg.user.email == "testing_user@testing.testing"
+    assert gpg.config.executable.endswith("gpg2")
+    assert gpg.user.passphrase == User._hash_passphrase(b"test_passphrase").hex()
+    assert gpg.user.email_address == "testing_user@testing.testing"
     assert gpg.fingerprint in gpg.list_keys()
     assert gpg.fingerprint in gpg.list_keys(secret=True)
     assert test_gpg.fingerprint in gpg.list_keys()
@@ -296,72 +298,90 @@ def test_export_import(gpg):
 
 
 def test_isolated_identities(gpg):
-    anon = GnuPG(
-        username="anon_user",
-        email="anonymous@testing.testing",
-        passphrase="rubbish_pw",
-        executable=_executable,
-    )
-    anon.gen_key()
-    anon_uid = anon.fingerprint
-    ### decrypting
-    enc_msg = gpg.encrypt("hi!", anon_uid)
-    msg = anon.decrypt(enc_msg)
-    try:
-        failed = False
-        gpg.decrypt(enc_msg)
-    except LookupError:
-        failed = True
-        """
-        Same user succefully prevented from decrypting message sent to a
-        different identity.
-        """
-    finally:
-        assert failed
-    ### signatures
-    sig_0 = anon.sign("message", local_user=anon_uid)
-    sig_1 = anon.sign("message")
-    try:
-        failed = False
-        gpg.sign("message", local_user=anon_uid)
-    except PermissionError:
-        failed = True
-        """
-        Same user successfully prevented from signing message with the
-        secret key associated with another identity.
-        """
-    finally:
-        assert failed
-    ### encrypting
-    try:
-        failed = False
-        gpg.encrypt("greetings", anon_uid, local_user=anon_uid)
-    except PermissionError:
-        failed = True
-        """
-        Same user successfully prevented from signing an encrypted message
-        with the secret key associated with another identity.
-        """
-    finally:
-        assert failed
-    while True:
+    """
+    Isolated identities do not work as originaly thought. The user must
+    choose unique passwords for each identity. Separating keys into
+    separate home directories helps only if .
+    """
+    with TemporaryDirectory("anon_user") as anon_homedir:
+        homedir = Path(anon_homedir).absolute()
+
+        user = User(
+            username="anon_user",
+            email_address="anonymous@testing.testing",
+            passphrase="test_passphrase",  # identities are isolated only if
+                                         # their passwords are NOT the same!
+        )
+        config = GnuPGConfig(homedir=str(homedir), executable=_executable)
+        anon = BaseGnuPG(user, config=config)
+
+        anon.generate_key()
+        anon_uid = anon.fingerprint
+        gpg.text_import(anon.text_export(anon_uid))
+        anon.text_import(gpg.text_export(gpg.fingerprint))
+        ### decrypting
+        enc_msg = gpg.encrypt("hi!", uid=anon_uid)
+        msg = anon.decrypt(enc_msg)
         try:
-            anon.delete(anon.email)
-        except:
-            break
+            failed = False
+            gpg.decrypt(enc_msg)
+        except LookupError:
+            failed = True
+            """
+            Same user succefully prevented from decrypting message sent to a
+            different identity.
+            """
+        finally:
+            assert failed
+        ### signatures
+        sig_0 = anon.sign("message", local_user=anon_uid)
+        sig_1 = anon.sign("message")
+        try:
+            failed = False
+            gpg.sign("message", local_user=anon_uid)
+        except PermissionError:
+            failed = True
+            """
+            Same user successfully prevented from signing message with the
+            secret key associated with another identity.
+            """
+        finally:
+            assert failed
+        ### encrypting
+        try:
+            failed = False
+            gpg.encrypt("greetings", uid=anon_uid, local_user=anon_uid)
+        except PermissionError:
+            failed = True
+            """
+            Same user successfully prevented from signing an encrypted message
+            with the secret key associated with another identity.
+            """
+        finally:
+            assert failed
+        while True:
+            try:
+                anon.delete(anon.user.email_address)
+            except:
+                break
 
 
 def test_cipher(gpg):
     test_email = "support@keys.openpgp.org"
     run(gpg.network_import(test_email))
     message = "\n  twenty\ntwo\narmed\ndogs\nrush\nthe\nkibble  \n\n"
-    for trust_level in range(0, 7):
+    for level in range(0, 7):
         for fingerprint in gpg.list_keys():
             try:
-                gpg.set_key_trust(fingerprint, trust_level)
-            except ValueError as invalid_trust_level:
+                failed = False
+                invalid_trust_level = (1 > level or level > 5)
+                gpg.set_key_trust(fingerprint, level)
+            except ValueError as error:
                 # if 1 > int(trust_level) or 5 < int(trust_level):
                 """Successfully blocked invlaid trust level"""
+                failed = True
+            finally:
+                assert failed if invalid_trust_level else not failed
         encrypted_message_0 = gpg.encrypt(
             message=message,
             uid=gpg.fingerprint,
@@ -410,16 +430,18 @@ def test_cipher(gpg):
         assert failed
     ###
     username = "test_sender"
-    email = "test_sender@testing.testing"
+    email_address = "test_sender@testing.testing"
     passphrase = "test_sender_passphrase"
 
-    sender = GnuPG(
-        email=email,
+    user = User(
+        email_address=email_address,
         username=username,
         passphrase=passphrase,
-        executable=_executable,
     )
-    sender.gen_key()
+    config = GnuPGConfig(homedir=_homedir, executable=_executable)
+    sender = BaseGnuPG(user, config=config)
+
+    sender.generate_key()
     sender_key = sender.list_keys(sender.fingerprint)
     sender_pkey = sender.text_export(sender.fingerprint)
     sender_skey = sender.text_export(sender.fingerprint, secret=True)
@@ -430,24 +452,18 @@ def test_cipher(gpg):
         failed = False
         gpg.decrypt(msg)
     except LookupError as warning:
-        try: gpg.list_keys(warning.value)
-        except: failed = True
+        failed = True
         gpg.text_import(sender_pkey)
-        gpg.list_keys(warning.value)
+        gpg.list_keys(warning.uid)
     finally:
-        while True:
-            try:
-                sender.delete(sender.fingerprint)
-            except:
-                break
         assert failed  # fingerprint is subkey of newly added key which
         # was derived from the returned ``decrypt()`` exception ``value``
         # attribute
 
 
 def test_file_io(gpg):
-    path = Path(gpg.homedir).absolute()
-    file_path = str(path / f"{gpg.fingerprint}.asc")
+    path = Path(gpg.config.homedir).absolute()
+    file_path = str(path / f"public-key_{gpg.fingerprint}.asc")
     key = gpg.text_export(gpg.fingerprint)
     gpg.file_export(path, gpg.fingerprint)
     gpg.file_import(file_path)
@@ -457,18 +473,18 @@ def test_file_io(gpg):
 def test_networking(gpg):
     dev_email = "gonzo.development@protonmail.ch"
     dev_fingerprint = "31FDCC4F9961AFAC522A9D41AE2B47FA1EF44F0A"
-    key_url = run(gpg.search(dev_email))
+    key_url = run(gpg.keyserver.search(dev_email))
     assert "\n" not in key_url
     assert " " not in key_url
     assert "<" not in key_url
     assert ">" not in key_url
-    key = run(gpg.network.get(key_url))
+    key = run(gpg.keyserver.network.get(key_url))
     gpg.text_import(key)
     assert gpg.list_keys(dev_email)
     fingerprint = gpg.key_fingerprint(dev_email)
     assert dev_fingerprint == fingerprint
     assert fingerprint in gpg.list_keys(dev_fingerprint)
-    email = gpg.key_email(fingerprint)
+    email = gpg.key_email_address(fingerprint)
     assert dev_email == email
     assert email == gpg.list_keys(dev_email)[fingerprint]
     key_from_email = gpg.text_export(email)
@@ -483,9 +499,9 @@ def test_networking(gpg):
     finally:
         assert len(gpg.list_keys(dev_email)) == 1
     run(gpg.network_export(gpg.fingerprint))
-    test_key_url = run(gpg.search(gpg.fingerprint))
+    test_key_url = run(gpg.keyserver.search(gpg.fingerprint))
     local_key = gpg.text_export(gpg.fingerprint)
-    network_key = run(gpg.network.get(test_key_url))
+    network_key = run(gpg.keyserver.network.get(test_key_url))
     assert local_key != network_key  # removed and/or reencoded uids
     try:
         failed = False
@@ -508,9 +524,13 @@ def test_networking(gpg):
     # "older" style header formatting, which leads to inconsistent
     # results when handling newer ECC keys.
     try:
+        failed = False
         run(gpg.network_import("nonsense uid data (HOPEFULLY)"))
     except FileNotFoundError:
+        failed = True
         """Successfully failed to retrieve data for bogus query"""
+    finally:
+        assert failed
 
 
 def test_network_concurrency(gpg):
@@ -521,11 +541,11 @@ def test_network_concurrency(gpg):
     async def looper(gpg, uid):
         tasks = []
         for i in range(2):
-            tasks.append(asyncio.ensure_future(gpg.search(uid)))
+            tasks.append(asyncio.ensure_future(gpg.keyserver.search(uid)))
         return tasks
 
     uid = "support@keys.openpgp.org"
-    url = run(gpg.search(uid))
+    url = run(gpg.keyserver.search(uid))
     urls = run(gather_looper(gpg, uid))
     assert url.strip()
     for link in urls:
@@ -546,7 +566,7 @@ def test_key_signing(gpg):
 
 
 def test_packet_parsing(gpg):
-    signature = gpg.sign("test")
+    signature = gpg.sign("testing")
     signed_encrypted_message = gpg.encrypt("test", gpg.fingerprint)
     encrypted_message = gpg.encrypt("test", gpg.fingerprint, sign=False)
     gpg_key = gpg.text_export(gpg.fingerprint)
@@ -611,9 +631,13 @@ def test_auto_fetch_methods(gpg):
     assert len(packets_3) == 28
     assert len(packets_4) == 31
     try:
+        failed = False
         gpg._list_packets(20 * "Non-OpenPGP data")
     except:
+        failed = True
         """Successfully failed when invalid data sent for parsing"""
+    finally:
+        assert failed
     ###
     fingerprint_0 = gpg._packet_fingerprint(dev_signed_encrypted_message)
     fingerprint_1 = gpg._packet_fingerprint(dev_encrypted_message)
@@ -640,8 +664,6 @@ def test_auto_fetch_methods(gpg):
         run(gpg.auto_verify(dev_signed_encrypted_message))
     except PermissionError as error:
         failed_correctly = True
-        notice = "``message`` is unverifiable."
-        assert notice in error.args[0]
     finally:
         assert failed_correctly  # signed encrypted message shows only
         # recipient from the outside (without the decryption key). Tester
@@ -656,8 +678,8 @@ def test_auto_fetch_methods(gpg):
         run(gpg.auto_verify(dev_encrypted_message))
     except Exception as exception:
         failed = True
-        keyid = exception.value
-        assert gpg.key_email(keyid) == gpg.key_email(keyserver_email)
+        keyid = exception.uid
+        assert gpg.key_email_address(keyid) == gpg.key_email_address(keyserver_email)
     finally:
         assert failed  # signed message shows only recipient from the
         # outside (without the decryption key).
@@ -692,29 +714,13 @@ def test_revoke(gpg):
 
 
 def test_delete(gpg):
+    test_fingerprint = gpg.fingerprint
     number_of_keys_before_delete = len(gpg.list_keys())
-    gpg.delete(gpg.fingerprint)
-    assert len(gpg.list_keys()) + 1 == number_of_keys_before_delete
-    while True:
-        try:
-            gpg.delete(gpg.email)
-        except:
-            break
-    while True:
-        try:
-            gpg.delete("gonzo.development@protonmail.ch")
-        except:
-            break
-    while True:
-        try:
-            gpg.delete("support@keys.openpgp.org")
-        except:
-            break
-    while True:
-        try:
-            gpg.delete("4826A9293FB8DF4765192455CDD760B5E60DB4F8")
-        except:
-            break
+
+    gpg.delete(test_fingerprint)
+    keyring_after_delete = gpg.list_keys()
+    assert test_fingerprint not in keyring_after_delete
+    assert number_of_keys_before_delete == 1 + len(keyring_after_delete)
 
 
 def test_reset_daemon(gpg):
